@@ -39,46 +39,12 @@
 #include <abz/sanitize.h>
 #include <abz/tokens.h>
 #include <abz/error.h>
+#include <abz/argv.h>
 #include <abz/run.h>
 
 #include <debug/memory.h>
 
-static char *tostr (const char *fmt,va_list ap)
-{
-   char *str;
-
-   if ((str = vbprintf (fmt,ap)) == NULL)
-	 {
-		abz_set_error ("failed to allocate memory: %m");
-		return (NULL);
-	 }
-
-   sanitize (str);
-
-   return (str);
-}
-
-static char **toargv (char *str)
-{
-   char **argv;
-   struct tokens tokens;
-
-   if (tokens_parse (&tokens,str))
-	 return (NULL);
-
-   if ((argv = mem_realloc (tokens.argv,sizeof (char *) * (tokens.argc + 1))) == NULL)
-	 {
-		abz_set_error ("failed to allocate memory: %m");
-		tokens_destroy (&tokens);
-		return (NULL);
-	 }
-
-   argv[tokens.argc] = NULL;
-
-   return (argv);
-}
-
-static __attribute__ ((noreturn)) void do_child (char **argv)
+static __attribute__ ((noreturn)) void do_child (char *const argv[])
 {
    int fd,result;
 
@@ -107,8 +73,9 @@ static __attribute__ ((noreturn)) void do_child (char **argv)
    _exit (127);
 }
 
-static int do_parent (pid_t pid,char *cmd,char *args)
+static int do_parent (pid_t pid,char *const argv[])
 {
+   char buf[256];
    int result;
 
    for (;;)
@@ -123,9 +90,10 @@ static int do_parent (pid_t pid,char *cmd,char *args)
 			 if (WEXITSTATUS (result))
 			   {
 				  abz_set_error ("child (%s) exited with non-zero exit status: %d",
-								 cmd,WEXITSTATUS (result));
+								 *argv,WEXITSTATUS (result));
 
-				  abz_cat_error ("attempted command: %s",args);
+				  argvtostr (buf,sizeof (buf),argv);
+				  abz_cat_error ("attempted command: %s",buf);
 
 				  return (-1);
 			   }
@@ -135,40 +103,32 @@ static int do_parent (pid_t pid,char *cmd,char *args)
 	 }
 }
 
-int vrun (const char *fmt,va_list ap)
+int runv (char *const argv[])
 {
-   size_t i;
-   char *str,**argv;
-   int result = -1;
    pid_t pid;
 
-   if ((str = tostr (fmt,ap)) == NULL)
-	 return (-1);
-
-   if ((argv = toargv (str)) == NULL)
+   if ((pid = fork ()) < 0)
 	 {
-		mem_free (str);
+		abz_set_error ("unable to fork process: %m");
 		return (-1);
 	 }
 
-   switch (pid = fork ())
-	 {
-	  case -1:
-		abz_set_error ("unable to fork process: %m");
-		result = -1;
-		break;
-	  case 0:
-		do_child (argv);
-	  default:
-		result = do_parent (pid,*argv,str);
-	 }
+   if (pid > 0)
+	 return (do_parent (pid,argv));
 
-   for (i = 0; argv[i] != NULL; i++)
-	 mem_free (argv[i]);
+   do_child (argv);
+}
 
-   mem_free (argv);
-   mem_free (str);
+int vrun (const char *fmt,va_list ap)
+{
+   int result;
+   char **argv;
 
+   if ((argv = vtoargv (fmt,ap)) == NULL)
+	 return (-1);
+   
+   result = runv (argv);
+   freeargv (argv);
    return (result);
 }
 
